# type: ignore[attr-defined]
"""Python API for acquisition and control of Edinburgh Sensors Gascard."""

import sys
from importlib import metadata as importlib_metadata


def get_version() -> str:
    """Get the version of the package.

    Returns:
        str: The version of the package.
    """
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()



"""Sets up the communication for the gas card device.

Author: Grayson Bellamy
Date: 2024-01-05
"""

from abc import ABC, abstractmethod
from collections.abc import ByteString

import anyio
import anyio.lowlevel
from anyserial import SerialStream
from anyserial.abstract import Parity, StopBits


class CommDevice(ABC):
    """Sets up the communication for the an Alicat device."""

    def __init__(self, timeout: int) -> None:
        """Initializes the serial communication.

        Args:
            timeout (int): The timeout of the Alicat device.
        """
        self.timeout = timeout

    @abstractmethod
    async def _read(self, len: int) -> ByteString | None:
        """Reads the serial communication.

        Args:
            len (int): The length of the serial communication to read. One character if not specified.

        Returns:
            str: The serial communication.
        """
        raise NotImplementedError

    @abstractmethod
    async def _write(self, command: str) -> None:
        """Writes the serial communication.

        Args:
            command (str): The serial communication.
        """
        raise NotImplementedError

    @abstractmethod
    async def close(self):
        """Closes the serial communication."""
        raise NotImplementedError

    @abstractmethod
    async def _readline(self) -> str | None:
        """Reads the serial communication until end-of-line character reached.

        Returns:
            str: The serial communication.
        """
        raise NotImplementedError

    @abstractmethod
    async def _write_readline(self, command: str) -> str | None:
        """Writes the serial communication and reads the response until end-of-line character reached.

        Args:
            command (str): The serial communication.

        Returns:
            str: The serial communication.
        """
        raise NotImplementedError


class SerialDevice(CommDevice):
    """Sets up the communication for the a gas card device using serial protocol."""

    def __init__(  # Need to verify Flow Control is set to 'none'
        self,
        port: str,
        baudrate: int = 57600,
        timeout: int = 150,  # Not present in manual
        databits: int = 8,
        parity: Parity = Parity.NONE,
        stopbits: StopBits = StopBits.ONE,
        xonxoff: bool = False,  # Not present in manual
        rtscts: bool = False,  # Not present in manual
        exclusive: bool = False,  # Not present in manual
    ):
        """Initializes the serial communication.

        Args:
            port (str): The port to which the Alicat device is connected.
            baudrate (int): The baudrate of the Alicat device.
            timeout (int): The timeout of the Alicat device in ms.
            databits (int): The number of data bits.
            parity (Parity): The parity of the Alicat device.
            stopbits (StopBits): The of stop bits. Usually 1 or 2.
            xonxoff (bool): Whether the port uses xonxoff.
            rtscts (bool): Whether the port uses rtscts.
            exclusive (bool): Whether the port is exclusive.
        """
        super().__init__(timeout)

        self.timeout = timeout
        self.eol = b"\r\n"
        self.serial_setup = {
            "port": port,
            "exclusive": exclusive,
            "baudrate": baudrate,
            "bytesize": databits,
            "parity": parity,
            "stopbits": stopbits,
            # "xonxoff": xonxoff,
            # "rtscts": rtscts,
        }
        self.isOpen = False
        self.ser_devc = SerialStream(**self.serial_setup)
        self.codes_list = [
            bytearray("N ", "ascii"),
            bytearray("N1 ", "ascii"),
            bytearray("C1 ", "ascii"),
            bytearray("E1 ", "ascii"),
            bytearray("O1 ", "ascii"),
            bytearray("X ", "ascii"),
            bytearray("U ", "ascii"),
        ]

    async def _read(self, len: int | None = None) -> ByteString | None:
        """Reads the serial communication.

        Args:
            len (int): The length of the serial communication to read. One character if not specified.

        Returns:
            ByteString: The serial communication.
        """
        if len is None:
            len = self.ser_devc.in_waiting()
            if len == 0:
                return None
        if not self.isOpen:
            async with self.ser_devc:
                return await self.ser_devc.receive_some(len)
        else:
            return await self.ser_devc.receive_some(len)
        return None

    async def _write(self, command: str) -> None:
        """Writes the serial communication.

        Note: Can writing ever actually timeout?

        Args:
            command (str): The serial communication.
        """
        if not self.isOpen:
            async with self.ser_devc:
                with anyio.move_on_after(self.timeout / 1000):
                    await self.ser_devc.send_all(command.encode("ascii") + self.eol)
        else:
            with anyio.move_on_after(self.timeout / 1000):
                await self.ser_devc.send_all(command.encode("ascii") + self.eol)
        return None

    async def _readline(self) -> str:
        """Reads the serial communication until end-of-line character reached.

        Returns:
            str: The serial communication.
        """
        async with self.ser_devc:
            self.isOpen = True
            line = bytearray()
            while True:  # Keep reading until end-of-line character reached, then we know new line is started
                c = None
                with anyio.move_on_after(
                    self.timeout / 1000
                ):  # if keep reading none, then timeout
                    while c is None:  # Keep reading until a character is read
                        c = await self._read()
                        await anyio.lowlevel.checkpoint()
                if c is None:  # if we reach timeout,
                    break
                line += c
                if line.startswith(
                    tuple(self.codes_list)
                ):  # This is a case where we started reading at the beginning of a new line
                    break
                if (
                    self.eol in line
                ):  # This is a case where we started reading in the middle of a line
                    line = bytearray()
                    break
            while True:
                if c is None:  # if we reach timeout, quit
                    break
                if self.eol in line:  # We already have a full line
                    break
                c = None
                with anyio.move_on_after(
                    self.timeout / 1000
                ):  # if keep reading none, then timeout
                    while c is None:  # Keep reading until a character is read
                        c = await self._read()
                        await anyio.lowlevel.checkpoint()
                line += c
        self.isOpen = False
        return line.decode("ascii")

    async def _write_readline(self, command: str) -> str:
        """Writes the serial communication and reads the response until end-of-line character reached.

        Parameters:
            command (str): The serial communication.

        Returns:
            str: The serial communication.
        """
        async with self.ser_devc:
            self.isOpen = True
            await self._write(command)
            line = bytearray()
            while True:  # Keep reading until end-of-line character reached, then we know new line is started
                c = None
                with anyio.move_on_after(
                    self.timeout / 1000
                ):  # if keep reading none, then timeout
                    while c is None:  # Keep reading until a character is read
                        c = await self._read()
                        await anyio.lowlevel.checkpoint()
                if c is None:  # if we reach timeout,
                    break
                line += c
                if line.startswith(
                    tuple(self.codes_list)
                ):  # This is a case where we started reading at the beginning of a new line
                    break
                if (
                    self.eol in line
                ):  # This is a case where we started reading in the middle of a line
                    line = bytearray()
                    break
            while True:
                if c is None:  # if we reach timeout, quit
                    break
                if self.eol in line:  # We already have a full line
                    break
                c = None
                with anyio.move_on_after(
                    self.timeout / 1000
                ):  # if keep reading none, then timeout
                    while c is None:  # Keep reading until a character is read
                        c = await self._read()
                        await anyio.lowlevel.checkpoint()
                line += c
        self.isOpen = False
        return line.decode("ascii")

    async def _flush(self) -> None:
        """Flushes the serial communication."""
        await self.ser_devc.discard_input()

    async def close(self) -> None:
        """Closes the serial communication."""
        self.isOpen = False
        await self.ser_devc.aclose()

    async def open(self) -> None:
        """Opens the serial communication."""
        self.isOpen = True
        await self.ser_devc.aopen()



"""Gascard device class.

Raises:
    ValueError: If device is not in the correct mode.
    ValueError: If device is not found on port.
    ValueError: If mode is invalid.
    ValueError: If gas is not accepted.

Returns:
    _type_: _description_
"""

import importlib.resources
import json
from abc import ABC
from typing import Any

from pygascard.comm import SerialDevice

codes_path = importlib.resources.files("pygascard").joinpath("codes.json")
with open(codes_path) as f:
    codes = json.load(f)
values = codes["values"]
N_labels = values["N"][0]
N1_labels = values["N1"][0]
C1_labels = values["C1"][0]
E1_labels = values["E1"][0]
O1_labels = values["O1"][0]
X_labels = values["X"][0]
U_labels = values["U"][0]


class Gascard(ABC):
    """Gascard class."""

    def __init__(
        self, device: SerialDevice, dev_info: dict[str, str], **kwargs: Any
    ) -> None:
        """Initialize the Gascard object.

        Args:
            device (SerialDevice): The serial device object.
            dev_info (dict): The device information dictionary.
            **kwargs: Additional keyword arguments.
        """
        self._device = device
        self._dev_info = dev_info
        self._current_mode = "U"
        self._MODES = (
            "N",
            "N1",
            "C1",
            "E1",
            "O1",
            "D",
            "X",
            "U",
        )

    @classmethod
    async def new_device(cls, port: str, **kwargs: Any) -> "Gascard":
        """Creates a new device. Chooses appropriate device based on characteristics.

        Example:
            dev = run(Gascard.new_device, "/dev/ttyUSB4")

        Args:
            port (str): The port of the device.
            **kwargs: Any

        Returns:
            Device: The new device.
        """
        if port.startswith("/dev/"):
            device = SerialDevice(port, **kwargs)
        await device._write("U")
        dev_info_raw = await device._readline()
        if not dev_info_raw:
            raise ValueError("No device found on port")
        dev_info_raw = dev_info_raw.replace("\x00", "")
        dev_info = dict(zip(U_labels, dev_info_raw.split()))
        if "U" not in dev_info_raw.split()[0]:
            # print("Error: Gas Card Not in User Interface Mode")
            raise ValueError("Gas Card Not in User Interface Mode")
        return cls(device, dev_info, **kwargs)

    async def _get_mode(self) -> str:
        """Gets the current mode of the device.

        Returns:
            str: Current mode of device
        """
        ret = await self._device._readline()
        mode = ret[:2].strip()
        if mode in self._MODES:
            self.current_mode = mode
        else:
            raise ValueError("Invalid Mode")
        return mode

    async def _set_mode(self, mode: str) -> None:
        """Sets the mode of the device.

        Args:
            mode (str): Desired mode for device
        """
        if mode in self._MODES:
            await self._device._write(mode)
            self._current_mode = mode
        else:
            raise ValueError("Invalid Mode")
        return

    async def _get_val(self) -> dict[str, str | float]:
        """Gets the current value of the device.

        Returns:
            dict[str, str | float]: Normal (N) mode dataframe
        """
        if self._current_mode != "N":
            await self._set_mode("N")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "N" not in df[0]:
            raise ValueError("Gas Card Not in Normal Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(N_labels, df))

    async def _get_raw(self) -> dict[str, str | float]:
        """Gets the raw sensor output.

        Returns:
            dict[str, str | float]: Normal Channel (N1) mode Dataframe
        """
        if self._current_mode != "N1":
            await self._set_mode("N1")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "N1" not in df[:2]:
            raise ValueError("Gas Card Not in Normal Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(N1_labels, df))

    async def _get_coeff(self) -> dict[str, str | float]:
        """Gets the current value of the device.

        Returns:
            dict[str, str | float]: Coefficient Channel (C1) mode dataframe
        """
        if self._current_mode != "C1":
            await self._set_mode("C1")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "C1" not in df[:2]:
            raise ValueError("Gas Card Not in Coefficient Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(C1_labels, df))

    async def _get_environmental(self) -> dict[str, str | float]:
        """Gets environmental parameters.

        Note:
            **WARNING Changing any environmental parameter will lead to incorrect gas sensor operation**

        Returns:
            dict: Environmental Mode (E1) dataframe
        """
        if self._current_mode != "E1":
            await self._set_mode("E1")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "E" not in df[0]:
            raise ValueError("Gas Card Not in Environmental Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(E1_labels, df))

    async def _get_output(self) -> dict[str, str | float]:
        """Display and Change output variables.

        Returns:
            dict[str, str | float]: Output Channel Mode (O1) dataframe
        """
        if self._current_mode != "O1":
            await self._set_mode("O1")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "O1" not in df[0:2]:
            raise ValueError("Gas Card Not in Output Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(O1_labels, df))

    async def _get_settings(self) -> dict[str, str | float]:
        """Display and Change Settings.

        Returns:
            dict[str, str | float]: Settings mode (X) dataframe
        """
        if self._current_mode != "X":
            await self._set_mode("X")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "X" not in df[0]:
            raise ValueError("Gas Card Not in Settings Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        return dict(zip(X_labels, df))

    async def _get_userinterface(self) -> dict[str, str | float]:
        """View user Interface.

        Returns:
            dict[str, str | float]: User Interface mode (U) dataframe
        """
        acc_gas = ["CO", "CO2", "CH4"]
        if self._current_mode != "U":
            await self._set_mode("U")
        ret = await self._device._readline()
        ret = ret.replace("\x00", "")
        df = ret.split()
        if "U" not in df[0]:
            raise ValueError("Gas Card Not in User Interface Mode")
        for index in range(len(df)):
            try:
                df[index] = float(df[index])
            except ValueError:
                pass
        if df[2] not in acc_gas:
            raise ValueError("Gas Not Accepted")
        return dict(zip(U_labels, df))

    async def get(self, vals: list[str] | None = None) -> dict[str, str | float]:
        """General function to receive from device.

        Max acquisition rate seems to be 4 Hz

        Example:
            df = run(dev.get, ["Gas Type", "Gas Range", "Conc 1"])
            df = run(dev.get, "Gas Type")

        Args:
            vals (list[str]): List of names (given in values dictionary) to receive from device.

        Returns:
            dict[str, str | float]: Dictionary of names requested with their values
        """
        if not vals:
            return await self._get_val()
        modes = []
        output = {}
        for val in vals:
            for key, value in values.items():
                for idx, names in enumerate(value[0]):
                    if val == names:
                        modes.append((key, names))
        unique_modes = {i[0] for i in modes}
        MODES_FUNC = {
            "N": self._get_val,
            "N1": self._get_raw,
            "C1": self._get_coeff,
            "E1": self._get_environmental,
            "O1": self._get_output,
            "X": self._get_settings,
            "U": self._get_userinterface,
        }
        for mode in unique_modes:
            ret = await MODES_FUNC[mode]()
            names = [i[1] for i in modes if i[0] == mode]
            output.update({names: ret[names] for names in names})
        return output

    async def set(self, params: dict[str, str | float]) -> None:
        """General function to send to device.

        Example:
            df = run(dev.set, {"Time Constant": 0, "Pressure Sensor Offset Cor": 904})

        Args:
            params (dict[str, str | float]): Variable:Value pairs for each desired set
        """
        modes = []
        for key, value in params.items():
            for key2, value2 in values.items():
                for idx, names in enumerate(value2[0]):
                    if key == names:
                        modes.append((key2, value2[1][idx], value))
        unique_modes = {i[0] for i in modes}
        for mode in unique_modes:
            if self._current_mode != mode:
                await self._set_mode(mode)
            for val in [i for i in modes if i[0] == mode]:
                await self._device._write(f"{val[1]}{val[2]}")
        return

    async def zero(self) -> None:
        """Sets the zero reference of the device.

        Example:
            df = run(dev.zero)

        Note:
            **Device MUST be flowing zero gas BEFORE calling this function.**
        """
        await self.set({"Zero_Gas_Corr_Factor": ""})
        return

    async def span(self, val: float) -> None:
        """Sets the span reference of the device.

        Example:
            df = run(dev.span, 0.2126)

        Note:
            **Device MUST be flowing span gas BEFORE calling this function.**

        Args:
            val (dict[str, float]): Gas concentration as a fraction of full scale (0.5 to 1.2)
        """
        await self.set({"Span_Gas_Corr_Factor": val})
        return

    async def time_const(self, val: int) -> None:
        """Sets the time constant of the RC filter of the device.

        Example:
            df = run(dev.time_const, 0)

        Args:
            val (int): Time constant in seconds (0 to 120)
        """
        await self.set({"Time_Constant": val})
        return
